#![allow(clippy::cast_possible_truncation, clippy::cast_precision_loss)]

pub(crate) mod bintokens;
pub(crate) mod bytes;
pub(crate) mod cfg;
pub(crate) mod lex;
pub(crate) mod recognizer;
pub(crate) mod rx;
pub(crate) mod svob;
pub(crate) mod toktree;
